// MOAR BARLS by LAch
// I'm supposed to be working on deltachars don't snowgrave me

if FFZHorizontalBarrels
	return
end

assert(ease, "FFZ Horizontal Barrels requires SRB2 v2.2.10 or above.")
assert(SPR_FFBA, "FFZ Horizontal Barrels requires FFZ Vertical Barrels loaded first.")

local HB = {}

rawset(_G, "FFZHorizontalBarrels", HB)

HB.CYCLE_TIME = 2*TICRATE // the time it takes for barrels to move from one waypoint to another
HB.HORIZONTAL_THRUST = 40*FRACUNIT // thrust power when firing from a barrel
HB.VERTICAL_THRUST = 2*FRACUNIT
HB.Ease = ease.inoutcubic // the easing function used to calculate barrel movement dynamics - see https://git.do.srb2.org/STJr/SRB2/-/merge_requests/1467 and https://easings.net for options

// ---
// SOC
// ---

freeslot("MT_FFZ_BARREL_HORIZONTAL_WAYPOINT", "MT_FFZ_BARREL_HORIZONTAL", "MT_FFZ_BARREL_HORIZONTAL_PART", "S_FFZ_BARREL_HORIZONTAL_PART")
FreeslotCarryType("CR_HORIZONTALBARREL")

mobjinfo[MT_FFZ_BARREL_HORIZONTAL_WAYPOINT] = {
	--$Name "Barrel Cannon (Horizontal) Waypoint"
	--$Sprite FFZHORIZ
	--$Category "Foliage Furnace"
	--$ParameterText "Tag"
	--$Flags1Text "Spawn barrel here"
	--$Flags2Text "Add 64 to tag"
	--$Flags4Text "Add 32 to tag"
	--$Flags8Text "Add 16 to tag"
	doomednum = 2362,
	spawnstate = S_INVISIBLE,
	radius = 64*FRACUNIT,
	height = 128*FRACUNIT,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

mobjinfo[MT_FFZ_BARREL_HORIZONTAL] = {
	spawnstate = S_INVISIBLE,
	radius = 64*FRACUNIT,
	height = 128*FRACUNIT,
	activesound = FFZ_BARREL_LATCH_SFX,
	painsound = FFZ_BARREL_FIRE_SFX,
	flags = MF_NOGRAVITY|MF_SPECIAL
}

mobjinfo[MT_FFZ_BARREL_HORIZONTAL_PART] = {
	spawnstate = S_FFZ_BARREL_HORIZONTAL_PART,
	radius = 80*FRACUNIT,
	height = 64*FRACUNIT,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

states[S_FFZ_BARREL_HORIZONTAL_PART] = {
	sprite = SPR_FFBA,
	frame = D | FF_PAPERSPRITE
}

// ---------
// functions
// ---------

local P_SetOrigin = P_SetOrigin or P_TeleportMove
local P_MoveOrigin = P_MoveOrigin or P_TeleportMove

// sets the origin of @slope without allocating temporary memory
local o = {}
function HB.SetSlopeOrigin(slope, x, y, z)
	o.x = x
	o.y = y
	o.z = z
	
	slope.o = o
end

// returns the tag of a barrel waypoint mapthing
function HB.GetTag(mt)
	local tag = mt.extrainfo
	local flags = mt.options
	
	if flags & MTF_OBJECTFLIP
		tag = $ + 64
	end
	if flags & MTF_OBJECTSPECIAL
		tag = $ + 32
	end
	if flags & MTF_AMBUSH
		tag = $ + 16
	end
	
	return tag
end

// spawns a barrel and all of its parts relative to the position and angle of @mo
function HB.CreateBarrel(mo)
	local angle = mo.angle
	local barrel = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_FFZ_BARREL_HORIZONTAL)
	local scale = barrel.scale
	local halfScale = barrel.scale >> 1
	local part = barrel
	
	local function CreatePart()
		part.hprev = P_SpawnMobjFromMobj(barrel, 0, 0, 0, MT_FFZ_BARREL_HORIZONTAL_PART)
		part.hprev.hnext = part
		part = $.hprev
	end
	
	barrel.tracer = mo
	barrel.angle = angle
	
	// create front and rear
	for i = 0, 1
		CreatePart()
		part.angle = angle + ANGLE_90
		part.movedir = i*ANGLE_180
		part.movefactor = 80*scale
		part.spriteyoffset = $ - 4*FRACUNIT
		part.z = $ + 64*scale
		part.frame = (B + i) | ($ & ~FF_FRAMEMASK)
	end
	
	// create left and right sides
	for i = 0, 1
		CreatePart()
		part.angle = angle
		part.movedir = ANGLE_90 + i*ANGLE_180
		part.movefactor = barrel.radius - halfScale
		part.spriteyscale = 32*FRACUNIT
		part.z = $ + 64*scale
	end
	
	// create top and bottom
	for i = 0, 1
		CreatePart()
		part.angle = angle
		part.spritexscale = 32*FRACUNIT
		part.spriteyoffset = $ - 85*FRACUNIT // I'm not sure why these need an extra frac
		part.z = $ + i*(barrel.height - scale) + halfScale
		part.renderflags = $ & ~RF_SPRITETYPEMASK | RF_FLOORSPRITE | RF_NOSPLATBILLBOARD
		part.frame = A
	end
	
	// create slopes
	for y = 0, 1
		for x = 0, 1
			local slope
			CreatePart()
			part.angle = angle
			part.movedir = ANGLE_90 + x*ANGLE_180
			part.movefactor = 47*scale + halfScale
			part.spritexscale = 16*FRACUNIT
			part.spriteyoffset = $ - 84*FRACUNIT
			part.z = $ + (barrel.height >> 1) + (2*y - 1) * (47*scale + halfScale)
			part.renderflags = $ & ~RF_SPRITETYPEMASK | RF_FLOORSPRITE | RF_NOSPLATBILLBOARD | RF_SLOPESPLAT | RF_NOSPLATROLLANGLE
			part.frame = A
			
			P_CreateFloorSpriteSlope(part)
			slope = part.floorspriteslope
			slope.xydirection = part.angle + part.movedir
			slope.zangle = ANGLE_45 - y*ANGLE_90
		end
	end
	
	HB.RotateTransformParts(barrel, 0, 0, 0, 0)
	
	return barrel
end

// moves @barrel's parts relative to its current position and angle
function HB.RotateTransformParts(barrel, xDelta, yDelta, zDelta, angle)
	local part = barrel.hprev
	
	while part
		local movefactor = part.movefactor
		local slope = part.floorspriteslope
		
		part.angle = $ + angle
		
		if movefactor
			local movedir = barrel.angle + part.movedir
			P_MoveOrigin(part,
				barrel.x + P_ReturnThrustX(part, movedir, movefactor),
				barrel.y + P_ReturnThrustY(part, movedir, movefactor),
				part.z + zDelta)
		else
			P_MoveOrigin(part, barrel.x, barrel.y, part.z + zDelta)
		end
		
		if slope
			slope.xydirection = $ + angle
			HB.SetSlopeOrigin(slope, part.x, part.y, part.z)
		end
		
		part = $.hprev
	end
end

// moves @barrel's parts relative to its position only
function HB.SlideTransformParts(barrel, xDelta, yDelta, zDelta)
	local part = barrel.hprev
	
	while part
		local slope = part.floorspriteslope
		
		P_MoveOrigin(part, part.x + xDelta, part.y + yDelta, part.z + zDelta)
		if slope
			HB.SetSlopeOrigin(slope, part.x, part.y, part.z)
		end
		
		part = $.hprev
	end
end

// moves a barrel to its position on its path
function HB.AdvanceBarrel(mo)
	local from = mo.tracer
	local to = mo.target
	
	local progress = HB.Ease(mo.threshold * FRACUNIT / HB.CYCLE_TIME, 0, FRACUNIT)
	
	local xDelta, yDelta, zDelta, angleDelta =
		from.x + FixedMul(to.x - from.x, progress) - mo.x,
		from.y + FixedMul(to.y - from.y, progress) - mo.y,
		from.z + FixedMul(to.z - from.z, progress) - mo.z,
		from.angle + FixedMul(to.angle - from.angle, progress) - mo.angle
	
	HB.TransformBarrel(mo, xDelta, yDelta, zDelta, angleDelta)
end

// moves a barrel relative to its current position and angle
function HB.TransformBarrel(mo, xDelta, yDelta, zDelta, angleDelta)
	P_MoveOrigin(mo, mo.x + xDelta, mo.y + yDelta, mo.z + zDelta)
	
	if angleDelta
		mo.angle = $ + angleDelta
		HB.RotateTransformParts(mo, xDelta, yDelta, zDelta, angleDelta)
	else
		HB.SlideTransformParts(mo, xDelta, yDelta, zDelta)
	end
	
	HB.UpdatePlayersInBarrel(mo)
end

// moves all players inside @barrel to its center and angles them
function HB.UpdatePlayersInBarrel(barrel)
	local mo = barrel
	local x, y, z = barrel.x, barrel.y, barrel.z + (barrel.height >> 1)
	
	while mo.hnext
		mo = $.hnext
		
		local player = mo.player
		
		if player.powers[pw_carry] ~= CR_HORIZONTALBARREL
			mo = $.hprev
			HB.UnlinkPlayer(player)
			continue
		end
		
		if player.pflags & PF_DIRECTIONCHAR
			player.drawangle = barrel.angle
		end
		player.pflags = $ | PF_STASIS | PF_THOKKED
		mo.momx, mo.momy, mo.momz = 0, 0, 0
		P_MoveOrigin(mo, x, y, z - (mo.height >> 1))
	end
end

// adds @player to @barrel, setting their pw_carry and playing sounds etc.
function HB.AddPlayerToBarrel(player, barrel)
	local mo = player.mo
	local tail = barrel
	
	while tail.hnext
		tail = $.hnext
	end
	
	mo.hprev = tail
	tail.hnext = mo
	
	P_ResetPlayer(player)
	player.pflags = $ | PF_STASIS | PF_THOKKED
	player.powers[pw_carry] = CR_HORIZONTALBARREL
	player.camerascale = $ << 1
	player.barrelstate = 0
	mo.tracer = barrel
	mo.state = S_PLAY_ROLL
	S_StartSound(mo, barrel.info.activesound)
	S_StartSound(mo, sfx_spin)
	
	mo.momx, mo.momy, mo.momz = 0, 0, 0
	P_SetOrigin(mo, barrel.x, barrel.y, barrel.z + (barrel.height - mo.height) >> 1)
end

// launches @player from their barrel, assumes they are in one
function HB.FirePlayer(player)
	local mo = player.mo
	local barrel = mo.tracer
	
	HB.ReleasePlayer(player)
	
	if not barrel
	or not barrel.hprev
		return
	end
	
	local angle = barrel.angle
	local radius = FixedDiv(barrel.hprev.radius, barrel.scale)
	local xDelta, yDelta = P_ReturnThrustX(mo, angle, radius), P_ReturnThrustY(mo, angle, radius)
	
	player.barrel = barrel // hijacking Rad's variables nyehehehe
	player.barrelstate = 2
	P_MoveOrigin(mo,
		barrel.x + FixedMul(xDelta, barrel.scale),
		barrel.y + FixedMul(yDelta, barrel.scale),
		barrel.z + (barrel.height - mo.height) >> 1)
	FFZ_SpawnBarrelExplosion(barrel, xDelta, yDelta, 32*FRACUNIT)
	P_SetObjectMomZ(mo, HB.VERTICAL_THRUST, false)
	P_InstaThrust(mo, angle, FixedMul(HB.HORIZONTAL_THRUST, barrel.scale))
	S_StartSound(mo, barrel.info.painsound)
end

// spins @player and removes them from their barrel, assumes they are in one
function HB.ReleasePlayer(player)
	P_ResetPlayer(player)
	player.pflags = $ | PF_SPINNING
	player.mo.state = S_PLAY_ROLL
	
	HB.UnlinkPlayer(player)
end

// removes @player from their barrel, assumes they are in one
function HB.UnlinkPlayer(player)
	local mo = player.realmo
	
	player.camerascale = $ >> 1
	
	if mo.hprev
		mo.hprev.hnext = mo.hnext
	end
	if mo.hnext
		mo.hnext.hprev = mo.hprev
	end
	
	mo.tracer = nil
	mo.hprev = nil
	mo.hnext = nil
end

// --------------
// hook functions
// --------------

function HB.TouchSpecialBarrel(barrel, mo)
	local player = mo.player
	local carry = player.powers[pw_carry]
	
	if not (carry == CR_NONE or carry == CR_PLAYER) // only allow barrel entry if you're not being carried, or if someone carries you into one
	or player.powers[pw_ignorelatch] & (1 << 15) // no entry if ignoring carrytypes
	or player.barrelstate == 1 // no entry if you're in a vertical barrel
	or player.barrelstate == 2 and player.barrel == barrel // no reentering the horizontal barrel you just fired from
		return
	end
	
	HB.AddPlayerToBarrel(player, barrel)
end

function HB.JumpSpecial(player)
	if player.pflags & PF_JUMPDOWN
	or player.powers[pw_carry] ~= CR_HORIZONTALBARREL
		return
	end
	
	HB.FirePlayer(player)
end

function HB.BarrelMobjThinker(mo)
	if mo.target == nil
		HB.UpdatePlayersInBarrel(mo)
		return // barrel has nowhere to go
	end
	
	mo.threshold = ($ + 1) % HB.CYCLE_TIME
	if mo.threshold == 0
		mo.tracer = mo.target 
		mo.target = mo.target.target
	end
	
	HB.AdvanceBarrel(mo)
end

function HB.BarrelMobjRemoved(barrel) // I didn't test this, I hope it works LMAO
	while barrel.hnext
		HB.ReleasePlayer(barrel.hnext.player)
	end
	
	while barrel.hprev
		P_RemoveMobj(barrel.hprev)
	end
end

function HB.WaypointMapThingSpawn(mo, mt)
	local type = mt.type
	local tag = HB.GetTag(mt)
	
	mo.flags = $ | MF_NOTHINK
	
	// spawn a barrel?
	if mt.options & MTF_EXTRA
		mo.angle = FixedAngle(mt.angle * FRACUNIT)
		mo.tracer = HB.CreateBarrel(mo)
	end
	
	// search through spawned mapthings to look for a link
	for i = 0, #mt - 1
		local mt = mapthings[i]
		
		if mt.type ~= type
		or HB.GetTag(mt) ~= tag
			continue
		end
		
		local mo2 = mt.mobj
		
		assert(mo2.target == nil, "More than 2 waypoints placed for horizontal barrel with tag " .. tag .. ".")
		
		mo.target = mo2
		mo2.target = mo
		
		// set barrel waypoints
		if mo.tracer
			mo.tracer.target = mo2
		end
		if mo2.tracer
			mo2.tracer.target = mo
		end
		break
	end
end

function HB.ShouldDamagePlayer(mo, _, _, _, damageType)
	local player = mo.player
	
	if not player
	or damageType & DMG_DEATHMASK
		return
	end
	
	if player.powers[pw_carry] == CR_HORIZONTALBARREL
		return false
	end
end

function HB.MapChange()
	for player in players.iterate
		if player.powers[pw_carry] == CR_HORIZONTALBARREL
			player.camerascale = $ >> 1
		end
	end
end

// -----
// hooks
// -----

addHook("MapThingSpawn", HB.WaypointMapThingSpawn, MT_FFZ_BARREL_HORIZONTAL_WAYPOINT)

addHook("MobjThinker", HB.BarrelMobjThinker, MT_FFZ_BARREL_HORIZONTAL)
addHook("MobjRemoved", HB.BarrelMobjRemoved, MT_FFZ_BARREL_HORIZONTAL)

addHook("TouchSpecial", function(...)
	HB.TouchSpecialBarrel(...)
	return true
end, MT_FFZ_BARREL_HORIZONTAL)

addHook("JumpSpecial", HB.JumpSpecial)

addHook("ShouldDamage", HB.ShouldDamagePlayer, MT_PLAYER)

addHook("MapChange", HB.MapChange)