-- Saw blade object thinkers by Tatsuru
local waypoints = {}
local sawblades = {}
local Lib = FFZLib

-- Saw blade setup.
addHook("MapLoad", function()
	waypoints = {}
	sawblades = {}
	
	for mapt in mapthings.iterate
		if mapt.type != mobjinfo[MT_SAWBLADEWP].doomednum continue end
		
		waypoints[mapt.angle] = $ or {}
		
		if waypoints[mapt.angle][mapt.extrainfo + 1]
			print("\x0082WARNING: \x0080Sawblade tag "..mapt.angle.." waypoint #"..(mapt.extrainfo + 1).." already exists.")
		end
		
		table.insert(waypoints[mapt.angle], mapt.extrainfo + 1, mapt.mobj)
	end
	
	for t, s in pairs(waypoints)
		local i = 1
		
		for o, m in ipairs(s)
			if i != o
				print("\x0082WARNING: \x0080Sawblade tag "..t.." waypoint #"..i.." is missing!")
				break
			end
			
			m.order = o
			m.target = s[o + 1] or s[1]
			i = $ + 1
		end
		
		if #s
			local saw = P_SpawnMobjFromMobj(s[1], 0, 0, 0, MT_SAWBLADE)
			saw.tag = t
			saw.scale = 2*FRACUNIT
			
			-- Did the mapper use a linedef to specify more settings?
			local l = lines[P_FindSpecialLineFromTag(800, t)]
		
			if l
				saw.customangle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
				saw.customspeed = l.frontside.textureoffset
				
				saw.waitfortrigger = l.flags & ML_BLOCKMONSTERS
				saw.stopatend = l.flags & ML_NOCLIMB
				saw.dieatend = l.flags & ML_EFFECT2
				saw.forceangle = l.flags & ML_BOUNCY
			end
			
			table.insert(sawblades, saw.tag, saw)
		end
	end
end)

-- Thinker for the saw blades.
addHook("MobjThinker", function(mo)
	-- Deactivating the helpers
	if not mo.pain and false
		mo.flags = $ & ~MF_PAIN
		
		for i = 0, 315, 45
			local pain = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SAWBLADEHELPER)
			pain.target = mo
			pain.rollangle = FixedAngle(i*FRACUNIT)
		end
		
		mo.pain = true
	end
	
	-- If we're stationary, we end here.
	if #waypoints[mo.tag] < 2
		S_StartSoundAtVolume(mo, sfx_sawbld, 90)
		return
	end
	
	-- Nothing to do but wait
	if mo.waitfortrigger
		mo.flags = $ & ~MF_PAIN
		mo.frame = $ & ~FF_ANIMATE
		return
	else
		mo.flags = $ | MF_PAIN
		mo.frame = $ | FF_ANIMATE
	end
	
	/*if P_IsObjectOnGround(mo) -- Commented out due to the obscene lag this is causing. Either we get rid of this or find a better more optimal solution to particles on saws.
		local spark = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_MINECARTSPARK)
		spark.momx = (P_RandomRange(-20,20) * mo.scale/10) + (mo.momx * -1) / 2
		spark.momy = (P_RandomRange(-20,20) * mo.scale/10) + (mo.momy * -1) / 2
		P_SetObjectMomZ(spark, P_RandomRange(0,20) * mo.scale/10, false)
		
		spark.scale = mo.scale/4
		spark.fuse = TICRATE/3
	end*/
	
	mo.target = $ or waypoints[mo.tag][2]

	if mo.customangle != nil and mo.forceangle
		mo.angle = mo.customangle
	else
		mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
	end
	
	local speed = mo.customspeed or mo.info.speed
	
	if Lib.HomingChase(mo, mo.target, speed)
		S_StartSoundAtVolume(mo, P_RandomChance(FRACUNIT/2) and sfx_sawim1 or sfx_sawim2, 70)
		
		if mo.target == waypoints[mo.tag][1]
			if mo.stopatend
				mo.waitfortrigger = true
			elseif mo.dieatend
				P_RemoveMobj(mo)
				return
			end
		end
		
		mo.target = mo.target.target
	end
	
	S_StartSoundAtVolume(mo, sfx_sawbld, 90)
end, MT_SAWBLADE)

-- Let the sawblade loose!
addHook("LinedefExecute", function(l)
	if not sawblades[l.tag] return end
	
	sawblades[l.tag].waitfortrigger = false
	--sawblades[l.tag] = nil
end, "RUNSAWBLADE")

local function SawCollide(mo, pmo)
	if not pmo.player return false end
	if pmo.player.playerstate != PST_LIVE return false end
	if pmo.player.powers[pw_flashing] return false end
	
	if Lib.HeightCheck(mo.z - mo.height*2, mo.z + mo.height*2, pmo.z, pmo.z + pmo.height)
		--Inherit the saw's momentum so it doesn't stop
		pmo.momx = $ + mo.momx
		pmo.momy = $ + mo.momy
		--pmo.momz = mo.momz
		
		return true
	end
end

/*
addHook("MobjCollide", SawCollide, MT_SAWBLADE)
addHook("MobjMoveCollide", SawCollide, MT_SAWBLADE)
*/

-- Thinker for the saw blade damage dealers.
addHook("MobjThinker", function(mo)
	if not mo.target
		P_RemoveMobj(mo)
		return
	end
	
	mo.flags = mo.target.waitfortrigger and ($ & ~MF_PAIN) or ($ | MF_PAIN)
	mo.angle = mo.target.angle
	
	if mo.target.waitfortrigger return end
	
	-- speeeeeen
	local ang = {0, ANG15, ANG30}
	local inc = ang[(leveltime%3) + 1]
	local sinex = FixedMul(mo.target.radius*2, cos(mo.rollangle + inc))
	
	local x = mo.target.x + FixedMul(sinex, cos(mo.angle))
	local y = mo.target.y + FixedMul(sinex, sin(mo.angle))
	local z = mo.target.z + FixedMul(mo.target.radius*2, sin(mo.rollangle + inc))
	
	P_TeleportMove(mo, x, y, z)
end, MT_SAWBLADEHELPER)

addHook("NetVars", function(net)
	waypoints = net($)
	sawblades = net($)
end)