--
--	Radicalicious' ATS launching barrels, made for Othius
--	and Kanna's Foliage Furnace Zone Act 2 in the Summer
--	2022 OLDC.
--

-- Freeslots
freeslot(
-- Sprites
"SPR_FFBA",

-- Object types
"MT_FFZ_BARREL_VERTICAL",
"MT_FFZ_BARREL_VERTICAL_WALL",
"MT_FFZ_BARREL_VERTICAL_BASE",

-- States
"S_FFZ_BARREL_VERTICAL",
"S_FFZ_BARREL_VERTICAL_WALL",
"S_FFZ_BARREL_VERTICAL_BASE_TOP",
"S_FFZ_BARREL_VERTICAL_BASE_BOTTOM"
)

rawset(_G, "FFZ_BARREL_LATCH_SFX", sfx_s3k4a) // Lach: just globalizing these for the horizontal barrels to use
rawset(_G, "FFZ_BARREL_FIRE_SFX", sfx_s3k81)
local FFZ_BARREL_SMOKE_SFX = sfx_s3kb4

local uncapped
if not P_SetOrigin then uncapped = false else uncapped = true end

-- Object definitions
mobjinfo[MT_FFZ_BARREL_VERTICAL] = {
		//$Name "Barrel Cannon (Vertical)"
		//$Sprite FFBAB0
		//$Category "Foliage Furnace"
		//$ParameterText "Speed"
		//$AngleText "FU to move"
		//$Flags2Text "Flip (broken, don't use!)"
		
		doomednum = 2361,
		spawnstate = S_FFZ_BARREL_VERTICAL,
		spawnhealth = 0,
		seestate = S_NULL,
		seesound = sfx_None,
		reactiontime = 0,
		attacksound = sfx_None,
		painstate = S_NULL,
		painchance = 0,
		painsound = sfx_None,
		meleestate = S_NULL,
		missilestate = S_NULL,
		deathstate = S_NULL,
		xdeathstate = S_NULL,
		deathsound = sfx_None,
		speed = 0,
		radius = 64*FRACUNIT,
		height = 160*FRACUNIT,
		dispoffset = 0,
		mass = 0,
		damage = 0,
		activesound = sfx_None,
		flags = MF_NOGRAVITY|MF_SPECIAL,
		raisestate = S_NULL
}

mobjinfo[MT_FFZ_BARREL_VERTICAL_WALL] = {
		doomednum = -1,
		spawnstate = S_FFZ_BARREL_VERTICAL_WALL,
		spawnhealth = 0,
		seestate = S_NULL,
		seesound = sfx_None,
		reactiontime = 0,
		attacksound = sfx_None,
		painstate = S_NULL,
		painchance = 0,
		painsound = sfx_None,
		meleestate = S_NULL,
		missilestate = S_NULL,
		deathstate = S_NULL,
		xdeathstate = S_NULL,
		deathsound = sfx_None,
		speed = 0,
		radius = 64*FRACUNIT,
		height = 160*FRACUNIT,
		dispoffset = 0,
		mass = 0,
		damage = 0,
		activesound = sfx_None,
		flags = MF_NOCLIP|MF_NOGRAVITY|MF_PAPERCOLLISION,
		raisestate = S_NULL
}

mobjinfo[MT_FFZ_BARREL_VERTICAL_BASE] = {
		doomednum = -1,
		spawnstate = S_FFZ_BARREL_VERTICAL_BASE_BOTTOM,
		spawnhealth = 0,
		seestate = S_NULL,
		seesound = sfx_None,
		reactiontime = 0,
		attacksound = sfx_None,
		painstate = S_NULL,
		painchance = 0,
		painsound = sfx_None,
		meleestate = S_NULL,
		missilestate = S_NULL,
		deathstate = S_NULL,
		xdeathstate = S_NULL,
		deathsound = sfx_None,
		speed = 0,
		radius = 64*FRACUNIT,
		height = FRACUNIT,
		dispoffset = 0,
		mass = 0,
		damage = 0,
		activesound = sfx_None,
		flags = MF_NOCLIP|MF_NOGRAVITY,
		raisestate = S_NULL
}

-- States
states[S_FFZ_BARREL_VERTICAL] = {
	sprite = SPR_NULL,
	frame = A,
	tics = -1,
	nextstate = S_NULL
}

states[S_FFZ_BARREL_VERTICAL_WALL] = {
	sprite = SPR_FFBA,
	frame = A|FF_PAPERSPRITE,
	tics = -1,
	nextstate = S_NULL
}

states[S_FFZ_BARREL_VERTICAL_BASE_TOP] = {
	sprite = SPR_FFBA,
	frame = B|FF_FLOORSPRITE,
	tics = -1,
	nextstate = S_NULL
}
states[S_FFZ_BARREL_VERTICAL_BASE_BOTTOM] = {
	sprite = SPR_FFBA,
	frame = C|FF_FLOORSPRITE,
	tics = -1,
	nextstate = S_NULL
}

-- Behavior
function A_TNTExplode(actor, var1, var2)
	if actor.harmless
		return
	else
		super(actor, var1, var2)
	end
end

local function P_DustRing(mobjtype, div, x, y, z, radius, speed, initscale, scale)
	local ang = FixedAngle(FixedDiv(360*FRACUNIT, div*FRACUNIT))

	-- it turned out the radius was effectively nullified thanks to errors in the original script
	-- BUT people preferred how it looked before I "fixed" it, so I got rid of the radius calculations altogether
	-- this was a bit of a mess to sort out, but at least it's probably somewhat fine now?
	-- -- Monster Iestyn (21/05/19)

	for i=0,div
		local dust = P_SpawnMobj(
			x, --+ FixedMul(radius, FINECOSINE((ang*i) >> ANGLETOFINESHIFT)),
			y, --+ FixedMul(radius, FINESINE((ang*i) >> ANGLETOFINESHIFT)),
			z,
			mobjtype
			)

		dust.angle = ang*i + ANGLE_90
		P_SetScale(dust, FixedMul(initscale, scale))
		dust.destscale = FixedMul(4*FRACUNIT + P_RandomFixed(), scale)
		dust.scalespeed = scale/24
		P_Thrust(dust, ang*i, speed + FixedMul(P_RandomFixed(), scale))
		dust.momz = P_SignedRandom()*scale/64
	end
end

rawset(_G, "FFZ_SpawnBarrelExplosion", function(mo, x, y, z)
	local explody = P_SpawnMobjFromMobj(mo, x, y, z, MT_THOK)
	explody.flags = $ | MF_NOCLIP
	explody.harmless = true
	explody.state = S_TNTBARREL_EXPL1
	explody.fuse = -1
	P_DustRing(MT_TNTDUST, 4, explody.x, explody.y, explody.z+explody.height, 64, 0, FRACUNIT, explody.scale)
	P_DustRing(MT_TNTDUST, 6, explody.x, explody.y, explody.z+explody.height/2, 96, FRACUNIT, FRACUNIT, explody.scale)
	return explody
end)

addHook("MobjSpawn", function(mo)
	local wall_1 = P_SpawnMobjFromMobj(mo, 0, 64*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_1.spritexscale = 32*FRACUNIT
	wall_1.angle = 0
	wall_1.target = mo
	local wall_2 = P_SpawnMobjFromMobj(mo, 64*FRACUNIT, 0, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_2.spritexscale = 32*FRACUNIT
	wall_2.angle = ANGLE_90
	wall_2.target = mo
	local wall_3 = P_SpawnMobjFromMobj(mo, 0, -64*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_3.spritexscale = 32*FRACUNIT
	wall_3.angle = ANGLE_180
	wall_3.target = mo
	local wall_4 = P_SpawnMobjFromMobj(mo, -64*FRACUNIT, 0, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_4.spritexscale = 32*FRACUNIT
	wall_4.angle = ANGLE_270
	wall_4.target = mo
	
	local wall_5 = P_SpawnMobjFromMobj(mo, 48*FRACUNIT, 48*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_5.spritexscale = 23*FRACUNIT
	wall_5.angle = ANGLE_135
	wall_5.target = mo
	local wall_6 = P_SpawnMobjFromMobj(mo, 48*FRACUNIT, -48*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_6.spritexscale = 23*FRACUNIT
	wall_6.angle = ANGLE_225
	wall_6.target = mo
	local wall_7 = P_SpawnMobjFromMobj(mo, -48*FRACUNIT, -48*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_7.spritexscale = 23*FRACUNIT
	wall_7.angle = ANGLE_315
	wall_7.target = mo
	local wall_8 = P_SpawnMobjFromMobj(mo, -48*FRACUNIT, 48*FRACUNIT, -4*FRACUNIT, MT_FFZ_BARREL_VERTICAL_WALL)
	wall_8.spritexscale = 23*FRACUNIT
	wall_8.angle = ANGLE_45
	wall_8.target = mo
	
	local base_1 = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_FFZ_BARREL_VERTICAL_BASE)
	base_1.renderflags = $ | RF_NOSPLATBILLBOARD
	base_1.angle = 0
	base_1.target = mo
	
	local base_2 = P_SpawnMobjFromMobj(mo, 0, 0, 160*FRACUNIT, MT_FFZ_BARREL_VERTICAL_BASE)
	base_2.state = S_FFZ_BARREL_VERTICAL_BASE_TOP
	base_2.renderflags = $ | RF_NOSPLATBILLBOARD
	base_2.angle = 0
	base_2.target = mo
	
	mo.origz = mo.z
end, MT_FFZ_BARREL_VERTICAL)

addHook("MapThingSpawn", function(mo, mt)
	mo.accelfactor = mt.extrainfo + 1
	mo.moveheight = mt.angle*FRACUNIT
end, MT_FFZ_BARREL_VERTICAL)

addHook("MobjThinker", function(mo)
	if not mo.valid then return end
	if not mo.target
		P_RemoveMobj(mo)
		return
	end
	if not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	local HRZOFFSET = 64*FRACUNIT
	local DIAOFFSET = 48*FRACUNIT
	local VRTOFFSET = -4*FRACUNIT
	
	if uncapped
		if mo.angle == 0
			P_MoveOrigin(mo, mo.target.x, mo.target.y+HRZOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_90
			P_MoveOrigin(mo, mo.target.x+HRZOFFSET, mo.target.y, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_180
			P_MoveOrigin(mo, mo.target.x, mo.target.y-HRZOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_270
			P_MoveOrigin(mo, mo.target.x-HRZOFFSET, mo.target.y, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_135
			P_MoveOrigin(mo, mo.target.x+DIAOFFSET, mo.target.y+DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_225
			P_MoveOrigin(mo, mo.target.x+DIAOFFSET, mo.target.y-DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_315
			P_MoveOrigin(mo, mo.target.x-DIAOFFSET, mo.target.y-DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_45
			P_MoveOrigin(mo, mo.target.x-DIAOFFSET, mo.target.y+DIAOFFSET, mo.target.z+VRTOFFSET)
		end
	else
		if mo.angle == 0
			P_TeleportMove(mo, mo.target.x, mo.target.y+HRZOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_90
			P_TeleportMove(mo, mo.target.x+HRZOFFSET, mo.target.y, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_180
			P_TeleportMove(mo, mo.target.x, mo.target.y-HRZOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_270
			P_TeleportMove(mo, mo.target.x-HRZOFFSET, mo.target.y, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_135
			P_TeleportMove(mo, mo.target.x+DIAOFFSET, mo.target.y+DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_225
			P_TeleportMove(mo, mo.target.x+DIAOFFSET, mo.target.y-DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_315
			P_TeleportMove(mo, mo.target.x-DIAOFFSET, mo.target.y-DIAOFFSET, mo.target.z+VRTOFFSET)
		elseif mo.angle == ANGLE_45
			P_TeleportMove(mo, mo.target.x-DIAOFFSET, mo.target.y+DIAOFFSET, mo.target.z+VRTOFFSET)
		end
	end
end, MT_FFZ_BARREL_VERTICAL_WALL)

addHook("MobjThinker", function(mo)
	if not mo then return end
	if not mo.valid then return end
	
	if mo.angle == 0 then return end
	if not mo.accelfactor then mo.accelfactor = 1 end
	
	mo.z = (mo.origz + mo.moveheight/2) + (sin(FixedAngle(mo.accelfactor*leveltime*FRACUNIT))*(mo.moveheight/2/FRACUNIT))
end, MT_FFZ_BARREL_VERTICAL)

addHook("MobjThinker", function(mo)
	if not mo.valid then return end
	if not mo.target
		P_RemoveMobj(mo)
		return
	end
	if not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	if uncapped
		if mo.state == S_FFZ_BARREL_VERTICAL_BASE_BOTTOM
			P_MoveOrigin(mo, mo.target.x, mo.target.y, mo.target.z)
		else
			P_MoveOrigin(mo, mo.target.x, mo.target.y, mo.target.z+160*FRACUNIT)
		end
	else
		if mo.state == S_FFZ_BARREL_VERTICAL_BASE_BOTTOM
			P_TeleportMove(mo, mo.target.x, mo.target.y, mo.target.z)
		else
			P_TeleportMove(mo, mo.target.x, mo.target.y, mo.target.z+160*FRACUNIT)
		end
	end
end, MT_FFZ_BARREL_VERTICAL_BASE)

addHook("TouchSpecial", function(barrel, pmo)
	if not barrel then return true end
	if not barrel.valid then return true end
	
	if not pmo then return true end
	if not pmo.valid then return true end
	if not pmo.player then return true end
	if not pmo.player.valid then return true end
	
	if pmo.player.barrelstate == nil then return true end
	if pmo.player.barrel == nil then return true end
	
	if pmo.player.barrelstate == 1
		return true
	else
		if (pmo.player.barrelstate == 2) and (pmo.player.barrel == barrel) and (pmo.momz > 0) then return true end
		S_StartSound(pmo, FFZ_BARREL_LATCH_SFX)
		S_StartSound(pmo, sfx_spin)
		pmo.player.barrelstate = 1
		pmo.player.barrel = barrel
		if pmo.player.cmd.buttons & BT_JUMP then pmo.player.barreljumplock = true end
	end
	return true
end, MT_FFZ_BARREL_VERTICAL)

addHook("PlayerSpawn", function(p)
	p.barrel = 0
	p.barrelstate = 0
end)

addHook("PlayerThink", function(p)
	if p.barrelstate == 1
		if uncapped
			P_SetOrigin(p.mo, p.barrel.x, p.barrel.y, p.barrel.z + 96*FRACUNIT)
		else
			P_TeleportMove(p.mo, p.barrel.x, p.barrel.y, p.barrel.z + 96*FRACUNIT)
		end
		p.mo.momx = 0
		p.mo.momy = 0
		p.mo.momz = 0
		p.mo.state = S_PLAY_ROLL
		p.pflags = $ | PF_STASIS | PF_THOKKED
		
		if p.cmd.buttons & BT_JUMP
			if not p.barreljumplock
				FFZ_SpawnBarrelExplosion(p.barrel, 0, 0, 160*FRACUNIT) // Lach: moved this into a function so horizontal barrels can use it
			
				p.mo.momx = 0
				p.mo.momy = 0
				p.mo.momz = 28*FRACUNIT
				S_StartSound(p.mo, FFZ_BARREL_FIRE_SFX)
				p.barrelstate = 2
			end
		else
			p.barreljumplock = false
		end
	elseif p.barrelstate == 2
		if leveltime % 6 == 0
			P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_SMOKE)
		end
		if leveltime % 24 == 0
			S_StartSound(p.mo, FFZ_BARREL_SMOKE_SFX)
		end
		
		if p.mo.momz > 0 then p.mo.momz = $ - FRACUNIT/4 end
		if p.mo.momz > -4*FRACUNIT then p.acceleration = skins[p.mo.skin].acceleration * 2
		else p.acceleration = skins[p.mo.skin].acceleration end
		
		if P_IsObjectOnGround(p.mo)
			p.acceleration = skins[p.mo.skin].acceleration
			p.barrelstate = 0
			p.barrel = 0
		end
	end
end)

addHook("MobjDamage", function(pmo, a, b, c, d)
	if not pmo then return end
	if not pmo.valid then return end
	if not pmo.player then return end
	if not pmo.player.valid then return end
	
	if pmo.player.barrelstate == 1
		return true
	else
		pmo.player.barrelstate = 0
		pmo.player.barrel = 0
		pmo.player.acceleration = skins[pmo.skin].acceleration
	end
end, MT_PLAYER)

addHook("MobjDeath", function(pmo, a, b, c, d)
	if not pmo then return end
	if not pmo.valid then return end
	if not pmo.player then return end
	if not pmo.player.valid then return end
	
	if pmo.player.barrelstate == 1
		return true
	else
		pmo.player.barrelstate = 0
		pmo.player.barrel = 0
		pmo.player.acceleration = skins[pmo.skin].acceleration
	end
end, MT_PLAYER)